import json

import numpy as np
import pandas as pd
import os

from dataman.tracker_utils import get_repo_info, in_ipynb, get_environment_yml
from .git_annex import AnnexRepo

_dont_check_git = False
_no_git_repo = False


def check_git_repos(when_dirty='warn'):
    import warnings
    _, is_dirty, script_repo = get_repo_info(os.path.dirname(track_info['entry_point']))

    if not _dont_check_git and is_dirty and not in_ipynb():
        if when_dirty == 'warn':
            warnings.warn("""Running from a dirty git repository (and not from a notebook).\n
         Please commit your changes before running. You will not be able to save data
                                 from this run""")
        else:
            raise RuntimeError("""Running from a dirty git repository (and not from a notebook).\n
                                  Please commit your changes before running""")

    if not _dont_check_git and is_dirty:
        d = script_repo.index.diff(None)
        if len(d) > 1:
            if when_dirty == 'warn':
                warnings.warn("""Running from a dirty git repo (besides the current notebook).
                                 Please commit your changes before running. \nYou will not be able to save data
                                 from this run""")
            else:
                raise RuntimeError("""Running from a dirty git repo (besides the current notebook).\n
                                      Please commit your changes before running""")

    print("len repos:")
    print(len(track_info['repos']))
    for repo in track_info['repos']:
        print(repo)
        _, is_dirty, _ = get_repo_info(repo['working_tree_dir'])
        if is_dirty:
            if when_dirty == 'warn':
                warnings.warn("Dependency repository " + repo['working_tree_dir'] + " is dirty, please commit!\n" +
                              "You will not  be able to save data from this run ")
            else:
                raise RuntimeError("Dependency repository " + repo['working_tree_dir'] + " is dirty, please commit!")


# noinspection PyProtectedMember
def _get_init_info():
    """This gets all the needed information about the run at the beginning.

    It is automatically called at module import
    Returns:
        a dict about the information
    """
    # this gets all the needed information at the beginning of the run
    info = {}

    # define a random UUID
    from uuid import uuid4
    info['uuid'] = str(uuid4())

    # report the time of run start
    from time import time
    info['run_time'] = str(time())

    # get name of the entry point and the arguments
    from sys import argv
    import os.path
    if in_ipynb():
        info['entry_point'] = '###notebook'
        info['args'] = []
    else:
        info['entry_point'] = os.path.realpath(argv[0])
        info['args'] = argv[1:]

    # get git status, if it's a script, this should be completely committed,
    # if it's a notebook everything should be committed
    # except for the notebook itself (which may be committed at the save time)
    repos = []

    if _no_git_repo:
        script_repo_info = {'working_tree_dir': ''}
    else:
        script_repo_info, is_dirty, script_repo = get_repo_info(os.path.dirname(info['entry_point']))

    repos.append(script_repo_info)

    # open config file, get git repos to be tracked, eventual files that need to be included in the dependencies
    # (e.g. lookup tables, etc.)
    # config file can be 1) in the current directory, named neuroseries.yml,
    # 2) in the 'project' directory (root of the containing git repo
    # 3) in the home directory as .neuroseries/config.yaml
    # 4) at the location pointed to by the variable NEUROSERIES_CONFIG

    config_candidates = ['./neuroseries.yml']

    import os.path
    config_candidates.append(os.path.join(repos[0]['working_tree_dir'], 'neuroseries.yml'))

    config_candidates.append('~/.neuroseries/config.yml')

    import os
    if 'NEUROSERIES_CONFIG' in os.environ:
        config_candidates.append(os.environ['NEUROSERIES_CONFIG'])

    import yaml
    config = {}
    for config_file in config_candidates:
        try:
            with open(os.path.expanduser(config_file)) as source:
                config = yaml.load(source)
                print('found config file at ' + config_file)
                break
        except FileNotFoundError:
            pass

    extra_repos = []
    info['config'] = config
    if 'extra_repos' in config:
        extra_repos.extend(config['extra_repos'])

    for r in extra_repos:
        script_repo_info, is_dirty, script_repo = get_repo_info(r)
        repos.append(script_repo_info)

    info['repos'] = repos

    # get venv status
    venv = get_environment_yml()
    info['venv'] = venv

    # get os information
    import platform
    os_info = dict(platform.uname()._asdict())
    info['os'] = os_info

    # get hardware information

    import psutil
    # noinspection PyProtectedMember
    mem_info = dict(psutil.virtual_memory()._asdict())
    info['memory'] = mem_info

    return info

track_info = _get_init_info()
check_git_repos('warn')

dependencies = []


def get_dependencies():
    return dependencies


def reset_dependencies():
    global dependencies
    dependencies = []


def str_to_series(a_string):
    """Helper function to convert strings
    This is mostly intended for json converted information, which is turned into a pandas series for storage in a
     pd.HDFStore. Only ASCII strings are supported

    Args:
        a_string: string to be converted

    Returns:
        a pd.Series
    """
    ba = bytearray(a_string, encoding='utf-8')
    ss = pd.Series(ba)
    return ss


def series_to_str(ss):
    """Helper function to convert a string encoded into a series

    Args:
        ss: a pd.Series encoded with str_to_series

    Returns: the decoded string

    """
    bb = bytearray(ss.values.tolist())
    bs = bytes(bb)
    a_string = bs.decode()
    return a_string


# noinspection PyClassHasNoInit
class PandasHDFStoreWithTracking(pd.HDFStore):
    """Enhanced pd.HDFStore including tracking information and metadata"""
    def get_info(self):
        """get information from file

        Returns: the information as a string

        """
        ss = self['file_info']
        return series_to_str(ss)

    def put_info(self, info):
        """
        Store info in the file
        Args:
            info: an information string

        Returns:
            None

        """
        ss = str_to_series(info)
        self.put('file_info', ss)

    def get_with_metadata(self, key):
        """get data with metadata

        Args:
            key: the variable name

        Returns:
            data: a pd.Series object (or nd.array)
            metadata: a dict with the metadata
        """
        data = self[key]
        metadata = None
        attr = self.get_storer(key).attrs
        if hasattr(attr, 'metadata'):
            metadata = attr.metadata

        return data, metadata

    def get(self, key):
        """Retrieves data from a store
        Pandas object and ndarray (via a pandas wrapper) are supported
        Args:
            key: the variable name

        Returns:
            data: the data

        """
        import ast
        data = super().get(key)
        attr = self.get_storer(key).attrs
        if attr is not None and hasattr(attr, 'metadata') and attr.metadata is not None and \
                        'class' in attr.metadata and attr.metadata['class'] == 'ndarray':
            data = data.values
            data_shape = ast.literal_eval(attr.metadata['shape'])
            data = data.reshape(data_shape)
        return data

    def put_with_metadata(self, key, value, metadata, **kwargs):
        """put data that needs ot be recorded with metadata

        Args:
            key: the variable name
            value: a pandas or ndarray
            metadata: a dict with the metadata
            **kwargs: additional arguments to be passed to store

        Returns:
            None

        """
        self.put(key, value, **kwargs)
        attr = self.get_storer(key).attrs

        if metadata is None:
            metadata = {}
        if hasattr(attr, 'metadata'):
            ex_metadata = attr.metadata
            metadata.update(ex_metadata)

        self.get_storer(key).attrs.metadata = metadata

    def put(self, key, value, **kwargs):
        """Put data in the store
        Handles pandas and ndarray types (the latter via a pandas wrapper)
        Args:
            key: the variable name
            value: a pandas or ndarray object
            **kwargs: optional arguments to be passed to pd.HDFStore

        Returns:
            None
        """
        if isinstance(value, np.ndarray):
            shape_str = str(value.shape)
            if value.ndim <= 2:
                value = pd.DataFrame(value)
            else:
                value = pd.Panel(value)
            super().put(key, value, **kwargs)
            metadata = {'class': 'ndarray', 'shape': shape_str}
            self.get_storer(key).attrs.metadata = metadata
        else:
            super().put(key, value, **kwargs)


class _GenericStore(object):
    def __init__(self, filename, mode):
        self.filename = filename
        self.mode = mode

    def get_info(self):
        pass

    def put_info(self, info):
        pass

    def close(self):
        pass


class FilesBackend(object):
    def fetch_file(self, filename):
        pass

    def repo_info(self, filename):
        return {'backend': 'FilesBackend', 'file': filename}

    def save_metadata(self, filename, info):
        pass

    def commit(self, filename, message_add=None):
        pass

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def get_file_info(self, filename):
        return None


default_backend = FilesBackend()


def set_default_backend(backend):
    global default_backend
    default_backend = backend


def get_hash_sha256(filename):
    import hashlib
    h256 = hashlib.sha256()

    f = open(filename, 'rb')

    for chunk in iter(lambda: f.read(4096), b""):
        h256.update(chunk)

    my_hash = h256.hexdigest()
    return my_hash


class JsonBackend(FilesBackend):
    def repo_info(self, filename):
        return {'backend': 'JsonBackend', 'file': filename}

    def save_metadata(self, filename, info):
        hash_file = get_hash_sha256(filename)
        info['hash'] = hash_file
        import os.path
        root, _ = os.path.splitext(filename)
        json_file = root + '.json'
        with open(json_file, 'w') as f:
            f.write(json.dumps(info))

    def get_file_info(self, filename):
        import os
        root, _ = os.path.splitext(filename)
        json_file = root + '.json'
        if os.path.exists(json_file):
            with open(json_file, 'r') as f:
                js_string = f.read()
            return json.loads(js_string)
        else:
            file_hash = self.get_hash('filename')
            info = {'file': filename, 'hash': file_hash}
            info.update(self.repo_info(filename))
            return info

    def get_hash(self, filename):
        return get_hash_sha256(filename)


class AnnexJsonBackend(JsonBackend):
    def __init__(self, dir_name=None, clone_from=None, new_repo=False, description=''):
        """Starts a new backend with git-annex support.
        Metadata are written in json files with the same name and '.json'
        extension. If new_repo is False, a git annex repo is searched for in:
        1) in dir_name
        2) in an 'annex' field of the config file
        3) in the current directory
        4) in current directory's parents
        Args:
            dir_name: The starting point for repo search
            clone_from: a URL containing a clonable repo
            new_repo: if True, a new repo will be created at dir_name or current directory
            description: a description to identify the current repo
        """
        import os
        import git
        if dir_name is None:
            if 'annex' in track_info['config']:
                dir_name = track_info['annex']
            else:
                dir_name = os.getcwd()
        dir_name = os.path.expanduser(dir_name)
        if not new_repo:  # then search for a git repo in parent dirs
            repo = git.Repo(dir_name, search_parent_directories=True)
            dir_name = repo.working_tree_dir
        self.repo = AnnexRepo(dir_name, clone_from=clone_from, description=description)

    def fetch_file(self, filename):
        """Attempts to ensure availability of file
        File will be looked for in remotes if not locally available
        Args:
            filename: the filename to be searched

        Returns: None

        """
        self.repo.get(filename)

    def repo_info(self, filename):

        """Get Backend-specific information

        Returns: a dict with the info

        """
        repo_info = {'backend': 'AnnexJsonBackend'}
        repo_info.update(self.repo.repo_info(filename))
        return repo_info

    def save_metadata(self, filename, info):
        """Saves file metadata in JSON file

        Args:
            filename: the filename the metadata is added for
            info: tracking information. The file hash is added here

        Returns:
            None

        """
        self.repo.add_annex(filename)
        hash_file = self.repo.lookupkey(filename)
        info['hash'] = hash_file
        import os.path
        root, _ = os.path.splitext(filename)
        json_file = root + '.json'
        with open(json_file, 'w') as f:
            f.write(json.dumps(info))
        self.repo.add(json_file)

    def commit(self, filename, message_add=None):
        """Commit the changes to git

        Args:
            filename: the file name to commit
            message_add: an additional message after the standard
            "Ran script.py. Added data.h5"

        Returns:
            None

        """
        import sys
        message = 'Ran ' + sys.argv[0] + '. Added ' + filename + '. '
        if message_add:
            message += message_add
        self.repo.commit(message)

    def get_hash(self, filename):
        return self.repo.lookupkey(filename)

    def push(self, remote_name):
        self.repo.push(remote_name)


def capture_info(item):
    import io
    from contextlib import redirect_stdout

    f = io.StringIO()
    with redirect_stdout(f):
        item.info()
    out = f.getvalue()
    return out


class TrackingStore(object):
    """A Store, encapsulating a file, that will keep track of code that has been run and its dependencies.
    What it does
    at construction
    1) fetch data wherever they are (backend)
    2) generate material store, material store should be able to handle pandas object (e.g. HDFStore)
    3) if read: get the tracking info if present add it to dependencies,
    if not generate file fingerprint, store that as tracking info
    4) if write: dump information
    file information is a dict {'file': filename, 'hash': an hash for the file, for example the SHA256 from git-annex,
    'date-created': a date string, 'run': track_info, 'dependencies': all the dependencies,
    'repo_info': information useful for the backend to retrieve the file
    'variables': a dict of variable names and info (as for example given by
    information is serialized as json and stored in the material store (backend)
    information also stored in json (text) metafile. before the file is closed,
    the hash will be the placeholder 'NULL'.
    it will be replaced in the metafile by the hash at closure

    material store must
    handle pandas objects, with metadata. Must be able to store (ASCII) info string
    must be able to write and read with or without metadata (for pandas)
    it can only do write and read, no append (enforced in this class)

    backend must
    fetch file (ensure availability)
    provide hash
    handle metafile info (or other mechanism). metafile info
    commit file and metafile (for example, at closure)
    for git-annex that would mean: data files in the annex, metafiles in the git repo #TODO

    backend is defined once per run and remains a property of run TODO
    """

    def __init__(self, filename, mode='r', backend=None, store_type=PandasHDFStoreWithTracking, comment=None,
                 **kwargs):
        """Initialize the Store

        Args:
            filename: The filename
            mode: can be 'r' (read) or 'w' (write)
            backend: the backend for file handling
            store_type: the class to be used for material store
            **kwargs: optional arguments, will be passed to the material store
        """
        check_git_repos(when_dirty='error')
        global dependencies
        if mode not in ['r', 'w']:
            raise ValueError('mode must be "w" or "r"')
        self.mode = mode
        self.filename = filename
        if backend:
            self.backend = backend
        else:
            self.backend = default_backend
        if mode == 'r':
            self.backend.fetch_file(filename)

        self.store = store_type(filename, mode, **kwargs)

        if mode == 'r':
            self.info = self.backend.get_file_info(self.filename)
            if self.info is None:  # if backend can't help, let's look into the store
                self.info = json.loads(self.store.get_info())

            if self.info is not None:
                dependencies.append(self.info)
        else:
            import datetime
            repo_info = self.backend.repo_info(self.filename)
            self.info = {'run': track_info, 'date_created': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                         'dependencies': dependencies, 'file': filename, 'hash': 'NULL', 'repo_info': repo_info,
                         'variables': {}}
            if comment:
                self.info['comment'] = comment

            self.store.put_info(json.dumps(self.info))
        print(dependencies)

    def close(self):
        """Close the store
        Does housekeeping with the tracking information and backend
        Returns:
            None
        """
        if self.mode == 'w':
            self.info['run'] = track_info
            self.store.put_info(json.dumps(self.info))
        self.store.close()
        self.backend.save_metadata(self.filename, self.info)
        self.backend.commit(self.filename)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get(self, key):
        """Gets data from store

        Args:
            key: the variable name

        Returns:
            the data
        """
        if self.mode == 'w':
            raise IOError("store is open in write mode")
        return self.store.get(key)

    def has_metadata(self, key):
        """checks if stored data has accompanying metadata

        Args:
            key: the variable name

        Returns:
            bool
        """
        if self.mode == 'w':
            raise IOError("store is open in write mode")
        (_, metadata) = self.store.get_with_metadata(key)
        return metadata is not None

    def get_with_metadata(self, key):
        """Get data with metadata

        Args:
            key: the variable name

        Returns:
            data: the data
            metadata: dict with metadata

        """
        if self.mode == 'w':
            raise IOError("store is open in write mode")
        return self.store.get_with_metadata(key)

    def keys(self):
        """lists the variables stored

        Returns:
            list of variable names

        """
        return self.store.keys()

    def put(self, key, value, metadata=None, **kwargs):
        """Put data in store

        Args:
            key: the variable name
            value: data
            metadata: dict with metadata
            **kwargs: optional arguments to be passed to material store

        Returns:
            None

        """
        if self.mode == 'r':
            raise IOError('store is open in read mode')
        self.info['variables'][key] = self.get_variable_info(key, value)

        self.store.put_with_metadata(key, value, metadata, **kwargs)

    def __setitem__(self, key, value):
        self.put(key, value)

    def __getitem__(self, item):
        return self.get(item)

    def get_file_info(self):
        """Returns the tracking information about the file

        Returns:
            a dict with all the stored information

        """
        backend_info = self.backend.get_file_info(self.filename)
        if backend_info:
            return backend_info
        return json.loads(self.store.get_info())

    def append(self, key, value, **kwargs):
        """Appends data to existing variable (if store allows it)

        Args:
            key: the variable name
            value: data to append
            **kwargs: optional arguments for the material store

        Returns:
            None

        """
        if self.mode == 'r':
            raise IOError('store is open in read mode')
        self.info['variables'][key] = self.get_variable_info(key, value)
        self.store.append(key, value, **kwargs)

    def get_variable_info(self, key, var):
        """Get information about a variable stored in the store

        Args:
            key: the key to the variable in the store
            var: the variable value

        Returns:
            A string including a description of the variable as obtained from Pandas or at the very least the type and
            shape of the variable

        """
        info = "Object of class: " + type(var).__name__ + '\n'
        try:
            i = capture_info(self.store[key])
            if i:
                info += i
        except (AttributeError, KeyError):
            try:
                i = capture_info(var)
                if i:
                    info += i
            except AttributeError:
                try:
                    s = str(var.shape)
                    info += 'Shape: ' + s
                except AttributeError:
                    import warnings
                    warnings.warn("Cannot determine info for variable. ")
        return info


HDFStore = TrackingStore


def store_file(filename, backend=None, comment=None):
    # noinspection PyTypeChecker
    store = TrackingStore(filename, mode='w', store_type=_GenericStore,  backend=backend, comment=comment)
    store.close()


def register_input(filename, backend=None):
    # noinspection PyTypeChecker
    store = TrackingStore(filename, mode='r', store_type=_GenericStore, backend=backend)
    store.close()



import pandas as pd
import numpy as np
from warnings import warn
from .time_series import TimeUnits, Range, as_method, store


# noinspection PyAbstractClass
class IntervalSet(pd.DataFrame):
    """
    a :py:class:`pandas.DataFrame` representing a (irregular) set of time intervals in elapsed time,
    with relative operations
    """
    def __init__(self, start, end=None, time_units=None, expect_fix=False, **kwargs):
        """
        makes a interval_set.

        If start and end and not aligned, meaning that len(start) == len(end),
        end[i] > start[i] and start[i+1] > end[i], or start and end are not sorted,
        will try to "fix" the data by eliminating some of the start and end data point
        :param start: array containing the beginning of each interval. If end is None, then it is taken to be a
        DataFrame-like object containing both the start and the end columns.
        :param end: array containing the end of each interval
        :param expect_fix: if False, will give a warning when a fix is needed (default: False)
        """

        if end is None:
            df = pd.DataFrame(start)
            if 'start' not in df.columns or 'end' not in df.columns:
                raise ValueError('wrong columns')
            super().__init__(df, **kwargs)
            self.r_cache = None
            self._metadata = ['nts_class']
            self.nts_class = self.__class__.__name__
            return

        start = np.array(start)
        end = np.array(end)
        start = TimeUnits.format_timestamps(start.ravel(), time_units,
                                            give_warning=not expect_fix)
        end = TimeUnits.format_timestamps(end.ravel(), time_units,
                                          give_warning=not expect_fix)

        to_fix = False
        msg = ''
        if not (np.diff(start) > 0).all():
            msg = "start is not sorted"
            to_fix = True
        if not (np.diff(end) > 0).all():
            msg = "end is not sorted"
            to_fix = True
        if len(start) != len(end):
            msg = "start and end not of the same length"
            to_fix = True
        else:
            # noinspection PyUnresolvedReferences
            if (start > end).any():
                msg = "some ends precede the relative start"
                to_fix = True
            # noinspection PyUnresolvedReferences
            if (end[:-1] > start[1:]).any():
                msg = "some start precede the previous end"
                to_fix = True

        if to_fix and not expect_fix:
            warn(msg, UserWarning)

        if to_fix:
            start.sort()
            end.sort()
            mm = np.hstack((start, end))
            mz = np.hstack((np.zeros_like(start), np.ones_like(end)))
            mx = mm.argsort()
            mm = mm[mx]
            mz = mz[mx]
            good_ix = np.nonzero(np.diff(mz) == 1)[0]
            start = mm[good_ix]
            end = mm[good_ix+1]

        # super().__init__({'start': start, 'end': end}, **kwargs)
        # self = self[['start', 'end']]
        data = np.vstack((start, end)).T
        super().__init__(data=data, columns=('start', 'end'), **kwargs)
        self.r_cache = None
        self._metadata = ['nts_class']
        self.nts_class = self.__class__.__name__

    def time_span(self):
        """
        Time span of the interval set.

        :return:  an IntervalSet with a single interval encompassing the whole IntervalSet
        """
        s = self['start'][0]
        e = self['end'].iloc[-1]
        return IntervalSet(s, e)

    def tot_length(self, time_units=None):
        """
        Total elapsed time in the set.

        :param time_units: the time units to return the result in ('us' [default], 'ms', 's')
        :return: the total length
        """
        tot_l = (self['end'] - self['start']).astype(np.float64).sum()
        return TimeUnits.return_timestamps(tot_l, time_units)

    def intersect(self, *a):
        """
        set intersection of IntervalSet's

        :param a: the IntervalSet to intersect self with, or a tuple of
        :return: the intersection IntervalSet
        """

        i_sets = [self]
        i_sets.extend(a)
        n_sets = len(i_sets)
        time1 = [i_set['start'] for i_set in i_sets]
        time2 = [i_set['end'] for i_set in i_sets]
        time1.extend(time2)
        time = np.hstack(time1)

        start_end = np.hstack((np.ones(len(time)//2, dtype=np.int32),
                              -1 * np.ones(len(time)//2, dtype=np.int32)))

        df = pd.DataFrame({'time': time, 'start_end': start_end})
        df.sort_values(by='time', inplace=True)
        df.reset_index(inplace=True, drop=True)
        df['cumsum'] = df['start_end'].cumsum()
        # noinspection PyTypeChecker
        ix = np.nonzero(df['cumsum'] == n_sets)[0]
        start = df['time'][ix]
        # noinspection PyTypeChecker
        end = df['time'][ix+1]

        return IntervalSet(start, end)

    def union(self, *a):
        """
        set union of IntervalSet's

        :param a:  the IntervalSet to intersect self with, or a tuple of
        :return: the union IntervalSet
        """
        i_sets = [self]
        i_sets.extend(a)
        time = np.hstack([i_set['start'] for i_set in i_sets] +
                         [i_set['end'] for i_set in i_sets])

        start_end = np.hstack((np.ones(len(time)//2, dtype=np.int32),
                              -1 * np.ones(len(time)//2, dtype=np.int32)))

        df = pd.DataFrame({'time': time, 'start_end': start_end})
        df.sort_values(by='time', inplace=True)
        df.reset_index(inplace=True, drop=True)
        df['cumsum'] = df['start_end'].cumsum()
        # noinspection PyTypeChecker
        ix_stop = np.nonzero(df['cumsum'] == 0)[0]
        ix_start = np.hstack((0, ix_stop[:-1]+1))
        start = df['time'][ix_start]
        stop = df['time'][ix_stop]

        return IntervalSet(start, stop)

    def set_diff(self, a):
        """
        set difference of IntervalSet's

        :param a: the interval set to set-subtract from self
        :return: the difference IntervalSet
        """
        i_sets = (self, a)
        time = np.hstack([i_set['start'] for i_set in i_sets] +
                         [i_set['end'] for i_set in i_sets])
        start_end1 = np.hstack((np.ones(len(i_sets[0]), dtype=np.int32),
                                -1 * np.ones(len(i_sets[0]), dtype=np.int32)))
        start_end2 = np.hstack((-1 * np.ones(len(i_sets[1]), dtype=np.int32),
                                np.ones(len(i_sets[1]), dtype=np.int32)))
        start_end = np.hstack((start_end1, start_end2))
        df = pd.DataFrame({'time': time, 'start_end': start_end})
        df.sort_values(by='time', inplace=True)
        df.reset_index(inplace=True, drop=True)
        df['cumsum'] = df['start_end'].cumsum()
        # noinspection PyTypeChecker
        ix = np.nonzero(df['cumsum'] == 1)[0]
        start = df['time'][ix]
        # noinspection PyTypeChecker
        end = df['time'][ix+1]

        return IntervalSet(start, end)

    def in_interval(self, tsd):
        """
        finds out in which element of the interval set each point in a time series fits.

        NaNs for those
        that don't fit a interval
        :param tsd: the tsd to be binned
        :return: an array with the interval index labels for each time stamp (NaN) for timestamps not in
        IntervalSet.
        """
        bins = self.values.ravel()
        ix = np.array(pd.cut(tsd.index, bins, labels=np.arange(len(bins) - 1, dtype=np.int64)))
        ix[np.floor(ix / 2) * 2 != ix] = np.NaN
        ix = np.floor(ix/2)
        return ix

    def drop_short_intervals(self, threshold, time_units=None):
        """
        Drops the short intervals in the interval set.

        :param threshold: time threshold for "short" intervals
        :type threshold: numeric
        :param time_units: the time units for the threshold
        :type time_units: str
        :return: a copied IntervalSet with the dropped intervals
        :rtype: neuroseries.interval_set.IntervalSet
        """
        threshold = TimeUnits.format_timestamps(np.array((threshold,), dtype=np.int64).ravel(), time_units)[0]
        return self.ix[(self['end']-self['start']) > threshold]

    def as_units(self, units=None):
        """
        returns a DataFrame with time expressed in the desired unit
        :param units: us (s), ms, or s
        :return: DataFrame with adjusted times
        """

        data = self.values.copy()
        data = TimeUnits.return_timestamps(data, units)
        df = pd.DataFrame(data=data, columns=self.columns)

        return df

    def merge_close_intervals(self, threshold, time_units=None):
        """
        Merges intervals that are very close.

        :param threshold: time threshold for the closeness of the intervals
        :param time_units: time units for the threshold
        :return: a copied IntervalSet with merged intervals
        """
        if len(self) == 0:
            return IntervalSet(start=[], end=[])
        tsp = self.time_span()
        i1 = tsp.set_diff(self)
        i1 = i1.drop_short_intervals(threshold, time_units=time_units)
        return tsp.set_diff(i1)

    @property
    def _constructor(self):
        return IntervalSet

    @property
    def r(self):
        """
        Time restricted version of the IntervalSet. (property, read-only)

        when running under a :py:class:`neuroseries.Range` context manager
        Returns:  when running under a :py:class:`neuroseries.Range` context manager


        """
        if Range.interval is None:
            raise ValueError('no range interval set')
        if self.r_cache is None:
            self.r_cache = self.intersect(Range.interval)
            Range.cached_objects.append(self)

        return self.r_cache

    def invalidate_restrict_cache(self):
        self.r_cache = None

    # def __iter__(self):
    #     self.iter_r = self.iterrows()
    #     return self
    #
    # def __next__(self):
    #     n = next(self.iter_r)[1]
    #     return IntervalSet(n['start'], n['end'])


IntervalSet.store = as_method(store)



import inspect
import os
import os.path
import os.path
import sys
from subprocess import Popen, PIPE

import git

ROOT_PREFIX = None

repos = []

dependencies = []


def get_caller(back=1):
    """
    get the main parameters of a calling function
    :param back: how many frames back in the stack you should go
    :return: a dict with fields 'filename', 'lineno', and 'function'
    """
    frame_caller = inspect.stack()[back]
    caller = {'filename': frame_caller[1], 'lineno': frame_caller[2],
              'function': frame_caller[3]}
    return caller


def get_entry_point():
    """
    get the entry point of a python script
    :return: a dict with fields: 'entry_script': the filename of the entry script,
    'entry_args': the arguments list with which it was called
    """
    frame_entry = inspect.stack()[-1]
    import sys
    entry = {'entry_script': frame_entry[1], 'entry_args': sys.argv}
    return entry


def get_repo_info(dir_name):
    """
    find the repository that contains the given directory
    :param dir_name: directory to be examined
    :return: a dict with fields 'working_tree_dir': hte
    """

    try:
        repo = git.Repo(dir_name, search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError:
        raise ValueError("no repository at " + dir_name)
    is_dirty = repo.is_dirty(untracked_files=True)
    repo_info = {'working_tree_dir': repo.working_tree_dir, 'commit': repo.head.ref.commit.hexsha,
                 'remote': repo.remotes.origin.url}
    return repo_info, is_dirty, repo


def is_repo_notebook_clean():
    """
    Tells you whether the script repository is clean except for the current notebook
    Returns: True if the repo is clean

    """

    from dataman import track_info
    script_repo_info, is_dirty, script_repo = get_repo_info(os.path.dirname(track_info['entry_point']))
    if not is_dirty:
        return True

    d = script_repo.index.diff(None)
    if len(d) > 1:
        return False

    changed_file = d[0].a_path

    _, notebook_path = get_notebook_name()
    from dataman.data_manager import track_info
    working_tree = track_info['repos'][0]['working_tree_dir']
    notebook_name = os.path.relpath(notebook_path, working_tree)

    if changed_file != notebook_name:
        return False

    return True


def commit_notebook():
    """
    commit the notebook (e.g. prior to saving data)
    Returns:
        True if the repo is now clean
        False otherwise

    """
    from dataman import track_info
    script_repo_info, is_dirty, script_repo = get_repo_info(os.path.dirname(track_info['entry_point']))
    ix = script_repo.index
    notebook_name, _ = get_notebook_name()
    ix.add(notebook_name)
    ix.commit("committing " + notebook_name)
    return not script_repo.is_dirty()


def in_ipynb():
    """
    determines if we are running within a ipython notebook
    :return: True if we are running in a notebook, False otherwise
    """
    try:
        # noinspection PyUnresolvedReferences,PyUnusedLocal
        cfg = get_ipython().config
        # if cfg['IPKernelApp']['parent_appname'] == 'ipython-notebook':
        #     return True
        # else:
        #     return False
        return True
    except NameError:
        return False


def run_magic(magic_string="timeit abs(-42)"):
    """
    runs a ipython magic command
    :param magic_string: the magic command
    :return: None
    """
    from IPython import get_ipython
    ipython = get_ipython()
    if ipython:
        ipython.magic(magic_string)


def fetch_notebook_name():
    """
    Fetches the notebook name via javascript so that it is available to get_notebook_name from the next cell
    :return: None
    """
    from IPython import get_ipython
    ipython = get_ipython()
    if ipython:
        ipython.run_cell_magic('javascript', '',
                               """
var kernel = Jupyter.notebook.kernel;
var command = ["notebookPath = ",
    "'", window.document.body.dataset.notebookPath, "'" ].join('')
kernel.execute(command)
var command = ["notebookName = ",
    "'", window.document.body.dataset.notebookName, "'" ].join('')
kernel.execute(command)
""")


def get_notebook_name(prefix=None):
    """
    Obtains ipython notebook path and name.

    This is an ugly hack, with an unfriendly user interface, but it is
    apparently the best we can do since jupyter doesn't expose the notebook path other than via javascript from the
    browser, and that has its own limitations. In order for this to work properly, the user must:
    1. Start the notebook server (under unix as jupyter notebook) from your home directory. If you choose not to do so
    then you have to specify the directory from which the server was started as the prefix argument here.
    2. reload the notebook in the browser after renaming the notebook, otherwise you will get a non-updated name.
    3. in a *previous cell* call save_notebook or fetch_notebook_name. The notebook name won't be available to this
    function until the cell in which one of these two function has been called has completed execution.

    :param prefix: the directory from which the notebook server is started
    (by default, the root of the git repo)
    :return: notebook_name: the notebook file name notebook_path: the notebook path name
    """
    if not prefix:
        from dataman.data_manager import track_info
        prefix = track_info['repos'][0]['working_tree_dir']

    stack = inspect.stack()
    notebook_path = ''
    notebook_name = ''
    for back in range(1, len(stack)):
        frame_caller = stack[back]
        # print(frame_caller[0].f_locals.keys())
        try:
            notebook_name = frame_caller[0].f_locals['notebookName']
            notebook_path = frame_caller[0].f_locals['notebookPath']
            notebook_path = os.path.join(prefix, notebook_path)
            break
        except KeyError:
            pass

    if notebook_name == '':
        raise RuntimeError("""Could not retrieve notebook path.
    If you are running a notebook, you need to call fetch_notebook or save_notebook from a previous cell.""")

    return notebook_name, notebook_path


def save_notebook():
    """
    Forces saving of the ipython notebook, raises RunTimeError if not in notebook,
    Also fetches the notebook name
    :return: None
    """
    from IPython import get_ipython
    ipython = get_ipython()
    if ipython:
        ipython.run_cell_magic('javascript', '',
                               """IPython.notebook.save_notebook();""")
        fetch_notebook_name()
    else:
        raise RuntimeError("Apparently not running a notebook at the moment")


def call_conda(extra_args, abspath=True):
    """
    calls conda in the virtual env where we are running from
    :param extra_args: the arguments to the conda call
    :param abspath: if True runs from the absolute path
    :return: stdout, stderr streams with conda output
    """
    # call conda with the list of extra arguments, and return the tuple
    # stdout, stderr (adapted from conda-api code)

    if abspath:
        if sys.platform == 'win32':
            # python = join(ROOT_PREFIX, 'python.exe')
            # conda  = join(ROOT_PREFIX, 'Scripts', 'conda-script.py')
            # TODO must somehow specify the environment under win32
            raise NotImplementedError
        else:
            cmd_list = ['conda']
    else:  # just use whatever conda is on the path
        cmd_list = ['conda']

    cmd_list.extend(extra_args)

    env1 = os.environ.copy()
    if 'PYTHONEXECUTABLE' in env1:
        del env1['PYTHONEXECUTABLE']

    try:
        if abspath:
            p = Popen(' '.join(cmd_list), stdout=PIPE, stderr=PIPE, shell=True, env=env1)
        else:
            # unix_path = os.getenv('PATH')
            # p = Popen(cmd_list, stdout=PIPE, stderr=PIPE, env={'PYTHONPATH': '', 'PATH': unix_path})
            p = Popen(cmd_list, stdout=PIPE, stderr=PIPE, shell=True, env=env1)
    except OSError:
        raise Exception("could not invoke {}\n".format(extra_args))
    return p.communicate()


def get_environment_yml():
    """
    gets the current environment in a form that can be used by conda to replicate it
    :return: a string including the content of the environment.yml file
    """

    s_out, s_err = call_conda(('env', 'export'))
    import re
    if len(s_err) > 0 and not re.match(r"^DEPRECATION", s_err.decode()):
        raise RuntimeError("conda failed with error: " + s_err.decode())
    return s_out.decode()

if __name__ == '__main__':

    stdout, stderr = call_conda(('env', 'export'))
    print(stdout.decode('utf-8'))
    print(stderr)



import git


class MyProgressPrinter(git.RemoteProgress):
    def update(self, op_code, cur_count, max_count=None, message=''):
        print(op_code, cur_count, max_count, cur_count / (max_count or 100.0), message or "NO MESSAGE")


class AnnexRepo(object):
    def __init__(self, path=None, clone_from=None, description=''):
        if path is None:
            import os
            path = os.getcwd()
        self.working_tree = path
        if clone_from:
            self.repo = git.Repo.clone_from(clone_from, to_path=path, progress=MyProgressPrinter)
            self.git = self.repo.git
            cmd_list = ['git', 'annex', 'init']
            if description:
                cmd_list.append('"' + description + '"')
            self.git.execute(cmd_list)
        else:
            self.repo = git.Repo.init(path)
            self.git = self.repo.git
            cmd_list = ['git', 'annex', 'init']
            if description:
                cmd_list.append('"' + description + '"')
            self.git.execute(cmd_list)
        self.remotes = {}
        self.is_special = {}
        import collections
        self.special_urls = collections.defaultdict(lambda: 'None')
        self.special_type = {}

    def path_in_repo(self, filename):
        import os
        path = os.path.relpath(os.path.abspath(filename), self.repo.working_tree_dir)
        return path

    def add_annex(self, file):
        file = self.path_in_repo(file)
        cmd_list = ['git', 'annex', 'add', file]
        self.git.execute(cmd_list)

    def add(self, file):
        file = self.path_in_repo(file)
        self.repo.index.add([file])

    def commit(self, message):
        self.repo.index.commit(message)

    def add_remote(self, name, remote_url):
        self.remotes[name] = self.repo.create_remote(name, remote_url)
        self.is_special[name] = False
        origin = self.remotes[name]
        origin.fetch()
        self.repo.create_head('master', origin.refs.master).set_tracking_branch(origin.refs.master).checkout()
    # def add_special_remote(self, name, remote_url, remote_type='rsync', auto_enable=True, **kwargs):
    #     special_types = {'rsync': self.add_special_remote_rsync}
    #     try:
    #         special_types[remote_type](name, remote_url, auto_enable, **kwargs)
    #     except KeyError:
    #         raise NotImplementedError("remote type not implemented. Types implemented are: "
    #                                   + str(special_types.keys()))

    # noinspection PyUnusedLocal

    def pull(self, remote_name):
        self.remotes[remote_name].pull()

    def push(self, remote_name):
        self.remotes[remote_name].push()

    # noinspection PyUnusedLocal
    def add_special_remote_rsync(self, name, remote_url, auto_enable=True, **kwargs):
        cmd_list = ['git', 'annex', 'initremote', name, 'type=rsync', 'rsyncurl=' + remote_url, 'encryption=none']
        if auto_enable:
            cmd_list.append('autoenable=true')
        self.git.execute(cmd_list)
        self.remotes[name] = self.repo.remotes[name]
        self.is_special[name] = True
        self.special_urls[name] = remote_url
        self.special_type[name] = 'rsync'

    def get(self, file):
        file = self.path_in_repo(file)
        cmd_list = ['git', 'annex', 'get', file]
        self.git.execute(cmd_list)

    def drop(self, file):
        file = self.path_in_repo(file)
        cmd_list = ['git', 'annex', 'drop', file]
        self.git.execute(cmd_list)

    def push_annex(self, remote_name):
        cmd_list = ['git', 'annex', 'sync', '--content', '--no-pull', self.remotes[remote_name].name]
        self.git.execute(cmd_list)

    def lookupkey(self, filename):
        filename = self.path_in_repo(filename)
        cmd_list = ['git', 'annex', 'lookupkey', filename]
        key = self.git.execute(cmd_list)
        return key

    def repo_info(self, file):
        normal_remotes = [k for k in self.remotes.keys() if not self.is_special[k]]
        normal_remotes_info = {k: {'name': self.remotes[k].name, 'url': self.remotes[k].url} for k in normal_remotes}

        special_remotes = [k for k in self.remotes.keys() if self.is_special[k]]
        special_remotes_info = {k: {'name': self.remotes[k].name, 'url': self.special_urls[k],
                                    'type': self.special_type} for k in special_remotes}

        path = self.path_in_repo(file)
        info = {'working_tree_dir': self.repo.working_tree_dir, 'remotes': normal_remotes_info,
                'special_remotes': special_remotes_info, 'path': path}
        return info



# noinspection PyPackageRequirements
from ipywidgets import interact
# noinspection PyPackageRequirements
from bokeh.models import Range1d
# noinspection PyPackageRequirements
from bokeh.io import push_notebook, show
# noinspection PyPackageRequirements
from bokeh.charts import TimeSeries
# noinspection PyPackageRequirements
from bokeh.plotting import figure
# noinspection PyPackageRequirements
from bokeh.models import ColumnDataSource, BoxAnnotation
# noinspection PyUnresolvedReferences,PyPackageRequirements
from bokeh.palettes import Spectral11

import pandas as pd
import numpy as np
import warnings


def FrameViewer(df, title="EEG", ylabel='voltage (mV)', xlabel='time (s)'):

    x_min = df.index[0]
    x_max = df.index[-1]
    default_interval = 1
    ts = TimeSeries(
        df, title=title,
        legend='top_left', ylabel=ylabel,
        xlabel=xlabel, x_range=Range1d(x_min, x_min+default_interval),
        plot_height=400)

    # noinspection PyShadowingNames
    def update_in(x_min=x_min, x_interval=default_interval):
        ts.x_range.start = x_min
        ts.x_range.end = x_min + x_interval
        push_notebook()

    show(ts, notebook_handle=True)
    interact(update_in, x_min=(x_min, x_max, 0.5), x_interval=(0.2, 5, 0.2))


# noinspection PyPep8Naming
def FrameViewerLong(df, title="EEG", ylabel='voltage (mV)', xlabel=None, downsample=10, units='s',
                    spread=0, annotations=None):

    from oio.open_ephys_io import ContinuousFile, load_continuous_tsd, is_sequence
    from .interval_set import IntervalSet
    # from .time_series import TsdFrame
    from .time_series import TimeUnits as tu

    if xlabel is None:
        xlabel = 'Time (' + units + ')'
    max_time = 30

    if is_sequence(df):
        if isinstance(df[0], str):
            df = [ContinuousFile(f) for f in df]
        x_min_df = df[0].start_time(units=units)
        x_max_df = df[0].end_time(units=units)
    else:
        x_min_df = df.start_time(units=units)
        x_max_df = df.end_time(units=units)

    if isinstance(df, pd.DataFrame):
        df = df.as_units(units)
        df1 = df[df.index < x_min_df + max_time]  # TODO use restrict
    else:
        df1 = load_continuous_tsd(df, t_min=tu.format_timestamps(x_min_df, units=units),
                                  t_max=tu.format_timestamps(x_min_df + max_time, units=units),
                                  downsample=downsample).as_units(units)

    default_interval = 1
    plot = figure(plot_height=400, plot_width=700, title=title,
                  tools="crosshair,pan,reset,save,box_zoom", x_axis_label=xlabel,
                  y_axis_label=ylabel)

    lines = []
    l = 0
    t = df1.index

    spread_vec = np.arange(len(df1.columns)) * spread
    for ix, x in enumerate(df1.columns):
        dd = {'t': t, x: df1[x] + spread_vec[ix]}
        source = ColumnDataSource(data=dd)
        lines.append(plot.line('t', x, source=source,
                               line_color=Spectral11[l % len(Spectral11)]))
        l += 1

    if annotations:
        for col, i_set in annotations.items():
            i_set = IntervalSet(x_min_df, x_min_df + max_time, time_units=units)
            i_set = i_set.as_units(units)

            for l in i_set.iterrows():
                box = BoxAnnotation(left=l[1]['start'], right=l[1]['end'], fill_alpha=0.1, fill_color=col)
                plot.add_layout(box)

    # noinspection PyShadowingNames
    def update_in(x_min=x_min_df, x_interval=default_interval):
        nonlocal df1

        if df1.index[0] > x_min or df1.index[-1] < x_min + x_interval:
            if isinstance(df, pd.DataFrame):
                df1 = df[((x_min - max_time/2) < df.index) & (df.index < (x_min + max_time/2))]
            else:
                df1 = load_continuous_tsd(df, t_min=tu.format_timestamps(x_min-max_time/2, units=units),
                                          t_max=tu.format_timestamps(x_min + max_time/2, units=units),
                                          downsample=downsample).as_units(units)
            t = df1.index
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                for ix, x in enumerate(df1.columns):
                    lines[ix].data_source.data['t'] = t
                    lines[ix].data_source.data[x] = df1[x] + spread_vec[ix]
            push_notebook()
        plot.x_range.start = x_min
        plot.x_range.end = x_min + x_interval
        push_notebook()

    show(plot, notebook_handle=True)
    interact(update_in, x_min=(x_min_df, x_max_df, 0.5), x_interval=(0.2, 5, 0.2))



import pandas as pd
import numpy as np
from warnings import warn
from pandas.core.internals import SingleBlockManager, BlockManager
from pandas.core.base import AccessorProperty, PandasObject

use_pandas_metadata = True


class BaseMethod(PandasObject):
    """
    Abstract class for converting functions into methods.

    .. This is based on the :func:`~pandas.core.base.AccessorProperty` class.

    It is mostly used here so that the same function may be applied to Tsd, TsdFrame and other classes
    """

    def __init__(self, data):
        """
        Creates an Accessor Property.

        Args:
            data: the object to be wrapped
        """
        self._data = data

    @staticmethod
    def func(*args, **kwargs):
        """
        dummy implementation for the function that will be wrapped

        Args:
            *args: Whatever the function requires
            **kwargs: Whatever the function requires

        Returns:
            None

        Raises:
            NotImplementedError: Always

        """
        raise NotImplementedError('should have a function defined here.')

    # noinspection PyNoneFunctionAssignment
    def __call__(self, *args, **kwargs):
        """
        This is the caller method for the wrapper
        Args:
            *args: Whatever the function requires
            **kwargs: Whatever the function requires

        Returns:
            Whatever the function returns

        """
        if len(args) == 0:
            fc = self.func(self._data, **kwargs)
        else:
            fc = self.func(self._data, *args, **kwargs)
        return fc

    __doc__ = func.__doc__
    __call__.__doc__ = func.__doc__


def as_method(func):
    """
    Converts a function into a method

    Using the AccessorProperty mechanism, see :func:`~neuroseries.BaseMethod`
    :param func: a function
    :return: a method
    """
    name = func.__name__.capitalize() + 'Method'
    cls = type(name, (BaseMethod,), {'func': staticmethod(func),
                                     '__doc__': func.__doc__})
    ap = AccessorProperty(cls, cls)
    return ap


class Range:
    """
    A class defining a range to restrict analyses.

    This is used as a context manager, taking a :func:`~neuroseries.interval_set.IntervalSet` as an input.
    After that, all neuroseries objects will have a property r set, that will be restricted, for example:

    .. code:: python

        with nts.Range(range_interval):
            np.testing.assert_array_almost_equal_nulp(self.tsd.r.times(), tsd_r.times())
    """
    interval = None
    cached_objects = []

    def __init__(self, a, b=None, time_units=None):
        """
        Creates a Range object
        Args:
            a: an :ref:`IntervalSet` defining the range, or the lower bound of the range
            b : if defined (defaults to :ref:`None`), contains the upper bound of the range, as a number of a Series
            (or other object with an Index)
            time_units (str): a string defining the :ref:`TimeUnits` used to define the bounds.
        """
        if b:
            start = TimeUnits.format_timestamps(np.array((a,), dtype=np.int64).ravel(), time_units)
            end = TimeUnits.format_timestamps(np.array((b,), dtype=np.int64).ravel(), time_units)
            from neuroseries.interval_set import IntervalSet
            Range.interval = IntervalSet(start, end)
        else:
            Range.interval = a

    def __enter__(self):
        return Range.interval

    def __exit__(self, exc_type, exc_val, exc_tb):
        Range.interval = None
        for i in Range.cached_objects:
            i.invalidate_restrict_cache()
        self.cached_objects = []


class TimeUnits:
    """
    This class deals with conversion between different time units for all neuroseries objects.
    It also provides a context manager that tweaks the default time units to the supported units:
    - 'us': microseconds (overall default)
    - 'ms': milliseconds
    - 's': seconds

    The context manager is called as follows

    .. code:: python

        with nts.TimeUnits('ms'):
            t = self.tsd.times()

    """
    default_time_units = 'us'

    def __init__(self, units):
        TimeUnits.default_time_units = units

    def __enter__(self):
        return self.default_time_units

    def __exit__(self, exc_type, exc_val, exc_tb):
        TimeUnits.default_time_units = 'us'

    @staticmethod
    def format_timestamps(t, units=None, give_warning=True):
        """
        Converts numerical types to the type :func:`numpy.int64` that is used for the time index in neuroseries.

        Args:
            t: a vector (or scalar) of times
            units: the units in which times are given
            give_warning: if True, it will warn when the timestamps are not sored

        Returns:
            ts: times in standard neuroseries format
        """

        import numbers

        if not units:
            units = TimeUnits.default_time_units

        # get the times (or the index if pandas object, in int64 vector format

        if isinstance(t, BlockManager):
            t = pd.DataFrame(t, copy=True)

        if isinstance(t, (pd.Series, pd.DataFrame)):
            t = t.index.values.astype(np.int64)

        if isinstance(t, np.floating):
            t = t.round()

        if isinstance(t, (pd.Series, pd.DataFrame)):
            t = t.index

        if isinstance(t, numbers.Number):
            t = np.array((t,))

        t = t.astype(np.float64)
        if units == 'us':
            pass
        elif units == 'ms':
            t *= 1000
        elif units == 's':
            t *= 1000000
        else:
            raise ValueError('unrecognized time units type')

        # noinspection PyUnresolvedReferences,PyTypeChecker
        ts = t.astype(np.int64).reshape((len(t),))

        if not (np.diff(ts) >= 0).all():
            if give_warning:
                warn('timestamps are not sorted', UserWarning)
            ts.sort()
        return ts

    @staticmethod
    def return_timestamps(t, units=None):
        """
        package the times in the desired units
        Args:
            t: standard neuroseries times
            units: the desired units for the output

        Returns:
            ts: times in the desired format
        """
        if units is None:
            units = TimeUnits.default_time_units
        if units == 'us':
            return t
        elif units == 'ms':
            return t / 1000.
        elif units == 's':
            return t / 1.0e6
        else:
            raise ValueError('Unrecognized units')


def _get_restrict_method(align):
    if align in ('closest', 'nearest'):
        method = 'nearest'
    elif align in ('next', 'bfill', 'backfill'):
        method = 'bfill'
    elif align in ('prev', 'ffill', 'pad'):
        method = 'pad'
    else:
        raise ValueError('Unrecognized restrict align method')
    return method


class Tsd(pd.Series):
    """
    A subclass of :func:`pandas.Series` specialized for neurophysiology time series.

    Tsd provides standardized time representation, plus functions for restricting and realigning time series
    """

    def __init__(self, t, d=None, time_units=None, **kwargs):
        """
        Tsd Initializer.

        Args:
            t: an object transformable in a time series, or a :func:`~pandas.Series` equivalent (if d is None)
            d: the data in the time series
            time_units: the time units in which times are specified (has no effect if a Pandas object
            is provided as the first argument
            **kwargs: arguments that will be passed to the :func:`~pandas.Series` initializer.
        """
        if isinstance(t, (pd.Series, SingleBlockManager)):
            super().__init__(t, **kwargs)
        else:
            t = TimeUnits.format_timestamps(t, time_units)
            super().__init__(index=t, data=d, **kwargs)
        self.index.name = "Time (us)"
        self._metadata.append("nts_class")
        self.nts_class = self.__class__.__name__
        self.r_cache = None

    def times(self, units=None):
        """
        The times of the Tsd, returned as np.double in the desired time units

        Args:
            units: the desired time units

        Returns:
            ts: the times vector

        """
        return TimeUnits.return_timestamps(self.index.values.astype(np.float64), units)

    def as_series(self):
        """
        The Tsd as a :func:`pandas:pandas.Series` object

        Returns:
            ss: the series object

        """

        return pd.Series(self, copy=True)

    def as_units(self, units=None):
        """
        Returns a Series with time expressed in the desired unit.

        :param units: us, ms, or s
        :return: Series with adjusted times
        """
        ss = self.as_series()
        t = self.index.values
        t = TimeUnits.return_timestamps(t, units)
        ss.index = t
        units_str = units
        if not units_str:
            units_str = 'us'
        ss.index.name = "Time (" + units_str + ")"
        return ss

    def data(self):
        """
        The data in the Series object

        Returns: the data

        """
        return self.values

    def realign(self, t, align='closest'):
        """
        Provides a new Series only including the data points that are close to one time point in the t argument.

        Args:
            t: the aligning series, in numpy or pandas format
            align: the values accepted by :func:`pandas.Series.reindex` plus
            - next (similar to bfill)
            - prev (similar to ffill)
            - closest (similar to nearest)

        Returns:
            The realigned Tsd

        """
        method = _get_restrict_method(align)
        ix = TimeUnits.format_timestamps(t.index.values)

        rest_t = self.reindex(ix, method=method)
        return rest_t

    def restrict(self, iset, keep_labels=False):
        """
        Restricts the Tsd to a set of times delimited by a :func:`~neuroseries.interval_set.IntervalSet`

        Args:
            iset: the restricting interval set
            keep_labels:

        Returns:

        """
        ix = iset.in_interval(self)
        tsd_r = pd.DataFrame(self, copy=True)
        col = tsd_r.columns[0]
        tsd_r['interval'] = ix
        ix = ~np.isnan(ix)
        tsd_r = tsd_r[ix]
        if not keep_labels:
            s = tsd_r.iloc[:, col]
            return Tsd(s)
        return Tsd(tsd_r, copy=True)

    @property
    def r(self):
        if Range.interval is None:
            raise ValueError('no range interval set')
        if self.r_cache is None:
            self.r_cache = self.restrict(Range.interval)
            Range.cached_objects.append(self)

        return self.r_cache

    def invalidate_restrict_cache(self):
        self.r_cache = None

    @property
    def _constructor(self):
        return Tsd


# noinspection PyAbstractClass
class TsdFrame(pd.DataFrame):
    def __init__(self, t, d=None, time_units=None, **kwargs):
        if isinstance(t, (pd.DataFrame, SingleBlockManager, BlockManager)):
            super().__init__(t, **kwargs)
        else:
            t = TimeUnits.format_timestamps(t, time_units)
            super().__init__(index=t, data=d, **kwargs)
        self.index.name = "Time (us)"
        self._metadata.append("nts_class")
        self.nts_class = self.__class__.__name__
        self.r_cache = None

    def times(self, units=None):
        return TimeUnits.return_timestamps(self.index.values.astype(np.float64), units)

    def as_dataframe(self, copy=True):
        """
        :return: copy of the data in a DataFrame (strip Tsd class label)
        """
        return pd.DataFrame(self, copy=copy)

    def as_units(self, units=None):
        """
        returns a DataFrame with time expressed in the desired unit
        :param units: us (s), ms, or s
        :return: DataFrame with adjusted times
        """
        t = self.index.values.copy()
        t = TimeUnits.return_timestamps(t, units)
        df = pd.DataFrame(index=t, data=self.values)
        units_str = units
        if not units_str:
            units_str = 'us'
        df.index.name = "Time (" + units_str + ")"
        df.columns = self.columns.copy()
        return df

    def plot(self, units=None):
        """
        makes a plot with the units of choices
        Args:
            units: us (s), ms, or s

        Returns:
            None
        """

        dz = pd.DataFrame(index=self.times(units=units), data=self.values, columns=self.columns, copy=False)
        units_str = units
        if not units_str:
            units_str = 'us'
        dz.index.name = "Time (" + units_str + ")"
        dz.plot()

    def data(self):
        if len(self.columns) == 1:
            return self.values.ravel()
        return self.values

    def realign(self, t, align='closest'):
        method = _get_restrict_method(align)
        ix = TimeUnits.format_timestamps(t)

        rest_t = self.reindex(ix, method=method, columns=self.columns.values)
        return rest_t

    def restrict(self, iset, keep_labels=False):
        ix = iset.in_interval(self)
        tsd_r = pd.DataFrame(self, copy=True)
        tsd_r['interval'] = ix
        ix = ~np.isnan(ix)
        tsd_r = tsd_r[ix]
        if not keep_labels:
            del tsd_r['interval']
        return TsdFrame(tsd_r, copy=True)

    @property
    def _constructor(self):
        return TsdFrame

    @property
    def _constructor_sliced(self):
        return Tsd

    @property
    def r(self):
        if Range.interval is None:
            raise ValueError('no range interval set')
        if self.r_cache is None:
            self.r_cache = self.restrict(Range.interval)
            Range.cached_objects.append(self)

        return self.r_cache

    def invalidate_restrict_cache(self):
        self.r_cache = None


# noinspection PyAbstractClass
class Ts(Tsd):
    def __init__(self, t, time_units=None, **kwargs):
        super().__init__(t, None, time_units=time_units, **kwargs)
        self.nts_class = self.__class__.__name__


def gaps(data, min_gap, method='absolute'):
    """
    finds gaps in a tsd
    :param data: a Tsd/TsdFrame
    :param min_gap: the minimum gap that will be considered
    :param method: 'absolute': min gap is expressed in time (us), 'median',
    min_gap expressed in units of the median inter-sample event
    :return: an IntervalSet containing the gaps in the TSd
    """
    dt = np.diff(data.times(units='us'))

    if method == 'absolute':
        pass
    elif method == 'median':
        md = np.median(dt)
        min_gap *= md
    else:
        raise ValueError('unrecognized method')

    ix = np.where(dt > min_gap)
    t = data.times()
    st = t[ix] + 1
    en = t[(np.array(ix) + 1)] - 1
    from neuroseries.interval_set import IntervalSet
    return IntervalSet(st, en)


# noinspection PyTypeChecker
Tsd.gaps = as_method(gaps)
# noinspection PyTypeChecker
TsdFrame.gaps = as_method(gaps)


def start_time(data, units='us'):
    return data.times(units=units)[0]


# noinspection PyTypeChecker
Tsd.start_time = as_method(start_time)
# noinspection PyTypeChecker
TsdFrame.start_time = as_method(start_time)


def end_time(data, units='us'):
    return data.times(units=units)[-1]

# noinspection PyTypeChecker
Tsd.end_time = as_method(end_time)
# noinspection PyTypeChecker
TsdFrame.end_time = as_method(end_time)


def support(data, min_gap, method='absolute'):
    """
    find the smallest (to a min_gap resolution) IntervalSet containing all the times in the Tsd
    :param data: a Tsd/TsdFrame
    :param min_gap: the minimum gap that will be considered
    :param method: 'absolute': min gap is expressed in time (us), 'median',
    min_gap expressed in units of the median inter-sample event
    :return: an IntervalSet
    """

    here_gaps = gaps(data, min_gap, method=method)
    t = data.times('us')
    from neuroseries.interval_set import IntervalSet
    span = IntervalSet(t[0] - 1, t[-1] + 1)
    support_here = span.set_diff(here_gaps)
    return support_here


# noinspection PyTypeChecker
Tsd.support = as_method(support)
# noinspection PyTypeChecker
TsdFrame.support = as_method(support)


# noinspection PyUnusedLocal
def filter_time_series(data, columns=None):
    pass


def store(data, the_store, key, **kwargs):
    if isinstance(data, Tsd):
        data_to_store = data.as_series()
    else:
        data_to_store = pd.DataFrame(data)

    the_store[key] = data_to_store
    # noinspection PyProtectedMember
    metadata = {k: getattr(data, k) for k in data._metadata}
    the_store.put(key, data_to_store, metadata, **kwargs)


# noinspection PyTypeChecker
Tsd.store = as_method(store)
# noinspection PyTypeChecker
TsdFrame.store = as_method(store)


def extract_from(storer):
    from neuroseries.interval_set import IntervalSet
    ks = storer.keys()
    extractable_classes = [Ts, Tsd, TsdFrame, IntervalSet]
    extractable_classes_id = {c.__name__: c for c in extractable_classes}

    variables = {}
    for k in ks:
        k = k[1:]
        (v, metadata) = storer.get_with_metadata(k)
        if metadata is not None and \
                        'nts_class' in metadata and \
                        metadata['nts_class'] in extractable_classes_id:
            variables[k] = extractable_classes_id[metadata['nts_class']](v)

        if hasattr(v, 'nts_class') and v.nts_class in extractable_classes_id:
            variables[k] = extractable_classes_id[v.nts_class](v)
    return variables



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from codecs import open
from os import path
import os
import subprocess

config = {

}

current_path = os.getcwd()
try:
    GIT_VERSION = subprocess.check_output(["git", "describe", "--tags"]).strip().decode('utf-8')
    GIT_VERSION = GIT_VERSION.split('-')[0]
except subprocess.CalledProcessError as e:
    GIT_VERSION = "0.1"
os.chdir(current_path)

print(GIT_VERSION)

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(description='neuro time series handling utilities',
      long_description=long_description,
      url='https://github.com/MemDynLab/neuroseries',
      author='Francesco P. Battaglia',
      author_email='fpbattaglia@gmail.com',
      version=GIT_VERSION,
      license='GPLv3',
      install_requires=['numpy', 'pandas==0.19.2', 'scipy', 'IPython', 'psutil', 'gitpython'],
      packages=['neuroseries', 'dataman'],
      package_data={'resources': ['resources/*.*']},
      include_package_data=True,
      name='neuroseries',
      classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Scientific/Engineering :: Bio-Informatics'
      ]
      )
